@@ -1,13 +1,24 @@
 from __future__ import print_function
+import sys
+import threading
+import time
+import functools
+import itertools
+import collections
+try:
+    import collections.abc as c
+except ImportError:
+    c = collections
+from decorator import (decorator, decorate, FunctionMaker, contextmanager,
+                       dispatch_on, __version__)
 
 doc = r"""\
 The ``decorator`` module
-=============================================================
 
 :Author: Michele Simionato
 :E-mail: michele.simionato@gmail.com
 :Version: $VERSION ($DATE)
-:Supports: Python 2.6, 2.7, 3.0, 3.1, 3.2, 3.3, 3.4, 3.5
+:Supports: Python 2.6, 2.7, 3.0, 3.1, 3.2, 3.3, 3.4, 3.5, 3.6, 3.7
 :Download page: http://pypi.python.org/pypi/decorator/$VERSION
 :Installation: ``pip install decorator``
 :License: BSD license
@@ -22,7 +33,7 @@ kicking. It is used by several frameworks (IPython, scipy, authkit,
 pylons, pycuda, sugar, ...) and has been stable for a *long*
 time. It is your best option if you want to preserve the signature of
 decorated functions in a consistent way across Python
-releases. Version 4.0 is fully compatible with the past, except for
+releases. Version 4 is fully compatible with the past, except for
 one thing: support for Python 2.4 and 2.5 has been dropped. That
 decision made it possible to use a single code base both for Python
 2.X and Python 3.X. This is a *huge* bonus, since I could remove over
@@ -30,7 +41,7 @@ decision made it possible to use a single code base both for Python
 separate docs for Python 2 and Python 3 effectively stopped any
 development on the module for several years. Moreover, it is now
 trivial to distribute the module as an universal wheel_ since 2to3 is no more
-required. Since Python 2.5 has been released 9 years ago, I felt that
+required. Since Python 2.5 has been released ages ago (in 2006), I felt that
 it was reasonable to drop the support for it. If you need to support
 ancient versions of Python, stick with the decorator module version
 3.4.2.  The current version supports all Python releases from 2.6 up to 3.6.
@@ -42,9 +53,9 @@ What's New in version 4
 
 - **New documentation**
   There is now a single manual for all Python versions, so I took the
-  opportunity to overhaul the documentation. So, even if you are
-  a long-time user, you may want to revisit the docs, since several
-  examples have been improved.
+  opportunity to overhaul the documentation and to move it to readthedocs.org.
+  Even if you are a long-time user, you may want to revisit the docs, since
+  several examples have been improved.
 
 - **Packaging improvements**
   The code is now also available in wheel format. Integration with
@@ -57,18 +68,25 @@ What's New in version 4
   ``decorator(caller, func)``. The old functionality is now deprecated
   and no longer documented, but still available for now.
 
-- **New experimental feature**
+- **Multiple dispatch**
   The decorator module now includes an implementation of generic
   functions (sometimes called "multiple dispatch functions").
   The API is designed to mimic ``functools.singledispatch`` (added
   in Python 3.4), but the implementation is much simpler.
   Moreover, all decorators involved preserve the signature of the
   decorated functions. For now, this exists mostly to demonstrate
-  the power of the module. In the future it could be enhanced/optimized;
-  however, its API could change. (Such is the fate of
-  experimental features!) In any case, it is very short and compact
-  (less then 100 lines), so you can extract it for your own use.
-  Take it as food for thought.
+  the power of the module. In the future it could be enhanced/optimized.
+  In any case, it is very short and compact (less then 100 lines), so you
+  can extract it for your own use. Take it as food for thought.
+
+- **Python 3.5 coroutines**
+  From version 4.1 it is possible to decorate coroutines, i.e. functions
+  defined with the `async def` syntax, and to maintain the
+  `inspect.iscoroutinefunction` check working for the decorated function.
+
+- **Decorator factories**
+  From version 4.2 there is facility to define factories of decorators in
+  a simple way, a feature requested by the users since a long time.
 
 Usefulness of decorators
 ------------------------------------------------
@@ -156,7 +174,7 @@ Here I used the functools.update_wrapper_ utility, which was added
 in Python 2.5 to simplify the writing of decorators.
 (Previously, you needed to manually copy the function attributes
 ``__name__``, ``__doc__``, ``__module__``, and ``__dict__``
-to the decorated function by hand.)
+to the decorated function by hand).
 
 Here is an example of usage:
 
@@ -179,9 +197,9 @@ keyword arguments:
 
 .. code-block:: python
 
- >>> from decorator import getargspec  # akin to inspect.getargspec
- >>> print(getargspec(f1))
- ArgSpec(args=[], varargs='args', varkw='kw', defaults=None)
+ >>> from decorator import getfullargspec
+ >>> print(getfullargspec(f1))
+ FullArgSpec(args=[], varargs='args', varkw='kw', defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})
 
 This means that introspection tools (like ``pydoc``) will give false
 information about the signature of ``f1`` -- unless you are using
@@ -196,10 +214,9 @@ calling the function with more than one argument raises an error:
     ...
  TypeError: f1() takes exactly 1 positional argument (2 given)
 
-Notice that ``inspect.getargspec`` and ``inspect.getfullargspec``
-will give the wrong signature. This even occurs in Python 3.5,
-although both functions were deprecated in that release.
-
+Notice that ``inspect.getfullargspec``
+will give the wrong signature, even in the latest Python, i.e. version 3.6
+at the time of writing.
 
 The solution
 -----------------------------------------
@@ -253,8 +270,8 @@ The signature of ``heavy_computation`` is the one you would expect:
 
 .. code-block:: python
 
- >>> print(getargspec(heavy_computation))
- ArgSpec(args=[], varargs=None, varkw=None, defaults=None)
+ >>> print(getfullargspec(heavy_computation))
+ FullArgSpec(args=[], varargs=None, varkw=None, defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})
 
 A ``trace`` decorator
 ------------------------------------------------------
@@ -285,8 +302,8 @@ It is immediate to verify that ``f1`` works...
 
 .. code-block:: python
 
- >>> print(getargspec(f1))
- ArgSpec(args=['x'], varargs=None, varkw=None, defaults=None)
+ >>> print(getfullargspec(f1))
+ FullArgSpec(args=['x'], varargs=None, varkw=None, defaults=None, kwonlyargs=[], kwonlydefaults=None, annotations={})
 
 The decorator works with functions of any signature:
 
@@ -299,8 +316,8 @@ The decorator works with functions of any signature:
  >>> f(0, 3)
  calling f with args (0, 3, 2), {}
 
- >>> print(getargspec(f))
- ArgSpec(args=['x', 'y', 'z'], varargs='args', varkw='kw', defaults=(1, 2))
+ >>> print(getfullargspec(f))
+ FullArgSpec(args=['x', 'y', 'z'], varargs='args', varkw='kw', defaults=(1, 2), kwonlyargs=[], kwonlydefaults=None, annotations={})
 
 $FUNCTION_ANNOTATIONS
 
@@ -354,14 +371,50 @@ Here is an example of usage:
  >>> func()
  calling func with args (), {}
 
-``blocking``
+The `decorator` function can also be used to define factories of decorators,
+i.e. functions returning decorators. In general you can just write something
+like this:
+
+.. code-block:: python
+
+   def decfactory(param1, param2, ...):
+      def caller(f, *args, **kw):
+          return somefunc(f, param1, param2, .., *args, **kw)
+      return decorator(caller)
+
+This is fully general but requires an additional level of nesting. For this
+reason since version 4.2 there is a facility to build
+decorator factories by using a single caller with default arguments i.e.
+writing something like this:
+
+.. code-block:: python
+
+   def caller(f, param1=default1, param2=default2, ..., *args, **kw):
+       return somefunc(f, param1, param2, *args, **kw)
+   decfactory = decorator(caller)
+
+Notice that this simplified approach *only works with default arguments*,
+i.e. `param1`, `param2` etc must have known defaults. Thanks to this
+restriction, there exists an unique default decorator, i.e. the member
+of the family which uses the default values for all parameters. Such
+decorator can be written as ``decfactory()`` with no parameters specified;
+moreover, as a shortcut, it is also possible to elide the parenthesis,
+a feature much requested by the users. For years I have been opposite
+to this feature request, since having expliciti parenthesis to me is more clear
+and less magic; however once this feature entered in decorators of
+the Python standard library (I am referring to the `dataclass` decorator
+https://www.python.org/dev/peps/pep-0557/) I finally gave up.
+
+The example below will show how it works in practice.
+
+Decorator factories
 -------------------------------------------
 
 Sometimes one has to deal with blocking resources, such as ``stdin``.
 Sometimes it is better to receive a "busy" message than just blocking
 everything.
-This can be accomplished with a suitable family of decorators,
-where the parameter is the busy message:
+This can be accomplished with a suitable family of decorators (decorator
+factory), parameterize by a string, the busy message:
 
 $$blocking
 
@@ -391,6 +444,15 @@ available. For instance:
  >>> print(read_data())
  some data
 
+Decorator factories are most useful to framework builders. Here is an example
+that gives an idea of how you could manage permissions in a framework:
+
+$$Action
+
+where ``restricted`` is a decorator factory defined as follows
+
+$$restricted
+
 ``decorator(cls)``
 --------------------------------------------
 
@@ -567,7 +629,7 @@ Here is what happens:
   an instance of ``FunctionMaker`` is created with the attributes
   ``args``, ``varargs``, ``keywords``, and ``defaults``.
   (These mirror the return values of the standard library's
-  ``inspect.getargspec``.)
+  ``inspect.getfullargspec``.)
 
 - For each item in ``args`` (a list of strings of the names of all required
   arguments), an attribute ``arg0``, ``arg1``, ..., ``argN`` is also generated.
@@ -583,8 +645,8 @@ followed by a tuple of defaults:
 
  >>> f1 = FunctionMaker.create(
  ...     'f1(a, b)', 'f(a, b)', dict(f=f), addsource=True, defaults=(None,))
- >>> print(getargspec(f1))
- ArgSpec(args=['a', 'b'], varargs=None, varkw=None, defaults=(None,))
+ >>> print(getfullargspec(f1))
+ FullArgSpec(args=['a', 'b'], varargs=None, varkw=None, defaults=(None,), kwonlyargs=[], kwonlydefaults=None, annotations={})
 
 
 Getting the source code
@@ -613,8 +675,7 @@ $$example
  <BLANKLINE>
 
 (See bug report 1764286_ for an explanation of what is happening).
-Unfortunately the bug still exists in all versions of Python, except
-Python 3.5.
+Unfortunately the bug still exists in all versions of Python < 3.5.
 
 However, there is a workaround. The decorated function has the ``__wrapped__``
 attribute, pointing to the original function. The simplest way to get the
@@ -697,6 +758,76 @@ following:
 - returns a value without making a recursive call; or,
 - returns directly the result of a recursive call.
 
+Python 3.5 coroutines
+-----------------------
+
+I am personally not using Python 3.5 coroutines yet, because at work we are
+still maintaining compatibility with Python 2.7. However, some users requested
+support for coroutines and since version 4.1 the decorator module has it.
+You should consider the support experimental and kindly report issues if
+you find any.
+
+Here I will give a single example of usage. Suppose you want to log the moment
+a coroutine starts and the moment it stops for debugging purposes. You could
+write code like the following:
+
+.. code-block:: python
+
+ import time
+ import logging
+ from asyncio import get_event_loop, sleep, wait
+ from decorator import decorator
+
+ @decorator
+ async def log_start_stop(coro, *args, **kwargs):
+     logging.info('Starting %s%s', coro.__name__, args)
+     t0 = time.time()
+     await coro(*args, **kwargs)
+     dt = time.time() - t0
+     logging.info('Ending %s%s after %d seconds', coro.__name__, args, dt)
+
+ @log_start_stop
+ async def make_task(n):
+     for i in range(n):
+         await sleep(1)
+
+ if __name__ == '__main__':
+     logging.basicConfig(level=logging.INFO)
+     tasks = [make_task(3), make_task(2), make_task(1)]
+     get_event_loop().run_until_complete(wait(tasks))
+
+and you will get an output like this::
+
+ INFO:root:Starting make_task(1,)
+ INFO:root:Starting make_task(3,)
+ INFO:root:Starting make_task(2,)
+ INFO:root:Ending make_task(1,) after 1 seconds
+ INFO:root:Ending make_task(2,) after 2 seconds
+ INFO:root:Ending make_task(3,) after 3 seconds
+
+This may be handy if you have trouble understanding what it going on
+with a particularly complex chain of coroutines. With a single line you
+can decorate the troubling coroutine function, understand what happens, fix the
+issue and then remove the decorator (or keep it if continuous monitoring
+of the coroutines makes sense). Notice that
+``inspect.iscoroutinefunction(make_task)``
+will return the right answer (i.e. ``True``).
+
+It is also possible to define decorators converting coroutine functions
+into regular functions, such as the following:
+
+.. code-block:: python
+
+ @decorator
+ def coro_to_func(coro, *args, **kw):
+     "Convert a coroutine into a function"
+     return get_event_loop().run_until_complete(coro(*args, **kw))
+
+Notice the diffence: the caller in ``log_start_stop`` was a coroutine
+function and the associate decorator was converting coroutines->coroutines;
+the caller in ``coro_to_func`` is a regular function and converts
+coroutines -> functions.
+
 Multiple dispatch
 -------------------------------------------
 
@@ -739,7 +870,7 @@ You can perform the registration with a decorator:
 
 $$writefloat
 
-Now XMLWriter can serialize floats:
+Now ``XMLWriter`` can serialize floats:
 
 .. code-block:: python
 
@@ -761,11 +892,11 @@ I have added an ordinal to the Rock-Paper-Scissors classes to simplify
 the implementation. The idea is to define a generic function (``win(a,
 b)``) of two arguments corresponding to the *moves* of the first and
 second players. The *moves* are instances of the classes
-Rock, Paper, and Scissors.
+Rock, Paper, and Scissors:
 
-Paper wins over Rock;
-Scissors wins over Paper; and
-Rock wins over Scissors.
+- Paper wins over Rock
+- Scissors wins over Paper
+- Rock wins over Scissors
 
 The function will return +1 for a win, -1 for a loss, and 0 for parity.
 There are 9 combinations, but combinations with the same ordinal
@@ -1211,7 +1342,7 @@ Here is an example:
 
 In order to introspect functions with annotations, one needs the
 utility ``inspect.getfullargspec`` (introduced in Python 3, then
-deprecated in Python 3.5, in favor of ``inspect.signature``):
+deprecated in Python 3.5, then undeprecated in Python 3.6):
 
 .. code-block:: python
 
@@ -1244,16 +1375,6 @@ Another attribute copied from the original function is ``__qualname__``,
 the qualified name. This attribute was introduced in Python 3.3.
 """
 
-import sys
-import threading
-import time
-import functools
-import itertools
-import collections
-import collections as c
-from decorator import (decorator, decorate, FunctionMaker, contextmanager,
-                       dispatch_on, __version__)
-
 if sys.version < '3':
     function_annotations = ''
 
@@ -1358,20 +1479,19 @@ def memoize(f):
     return decorate(f, _memoize)
 
 
-def blocking(not_avail):
-    def _blocking(f, *args, **kw):
-        if not hasattr(f, "thread"):  # no thread running
-            def set_result():
-                f.result = f(*args, **kw)
-            f.thread = threading.Thread(None, set_result)
-            f.thread.start()
-            return not_avail
-        elif f.thread.isAlive():
-            return not_avail
-        else:  # the thread is ended, return the stored result
-            del f.thread
-            return f.result
-    return decorator(_blocking)
+@decorator
+def blocking(f, msg='blocking', *args, **kw):
+    if not hasattr(f, "thread"):  # no thread running
+        def set_result():
+            f.result = f(*args, **kw)
+        f.thread = threading.Thread(None, set_result)
+        f.thread.start()
+        return msg
+    elif f.thread.isAlive():
+        return msg
+    else:  # the thread is ended, return the stored result
+        del f.thread
+        return f.result
 
 
 class User(object):
@@ -1386,47 +1506,42 @@ class Admin(PowerUser):
     "Will be able to delete pages too"
 
 
-def get_userclass():
-    return User
-
-
 class PermissionError(Exception):
-    pass
-
-
-def restricted(user_class):
-    def restricted(func, *args, **kw):
-        "Restrict access to a given class of users"
-        userclass = get_userclass()
-        if issubclass(userclass, user_class):
-            return func(*args, **kw)
-        else:
-            raise PermissionError(
-                '%s does not have the permission to run %s!'
-                % (userclass.__name__, func.__name__))
-    return decorator(restricted)
-
-
-class Action(object):
     """
     >>> a = Action()
+    >>> a.user = User()
     >>> a.view() # ok
     >>> a.insert() # doctest: +IGNORE_EXCEPTION_DETAIL
     Traceback (most recent call last):
        ...
     PermissionError: User does not have the permission to run insert!
     """
+
+
+@decorator
+def restricted(func, user_class=User, *args, **kw):
+    "Restrict access to a given class of users"
+    self = args[0]
+    if isinstance(self.user, user_class):
+        return func(*args, **kw)
+    else:
+        raise PermissionError(
+            '%s does not have the permission to run %s!'
+            % (self.user, func.__name__))
+
+
+class Action(object):
     @restricted(User)
     def view(self):
-        pass
+        "Any user can view objects"
 
     @restricted(PowerUser)
     def insert(self):
-        pass
+        "Only power users can insert objects"
 
     @restricted(Admin)
     def delete(self):
-        pass
+        "Only the admin can delete objects"
 
 
 class TailRecursive(object):
