@@ -6,10 +6,15 @@ import decimal
 import inspect
 import functools
 import collections
+from collections import defaultdict
+try:
+    c = collections.abc
+except AttributeError:
+    c = collections
 from decorator import dispatch_on, contextmanager, decorator
 try:
     from . import documentation as doc
-except (SystemError, ValueError):
+except (ImportError, ValueError, SystemError):  # depending on the py-version
     import documentation as doc
 
 
@@ -24,6 +29,35 @@ def assertRaises(etype):
         raise Exception('Expected %s' % etype.__name__)
 
 
+if sys.version >= '3.5':
+    exec('''from asyncio import get_event_loop
+
+@decorator
+async def before_after(coro, *args, **kwargs):
+    return "<before>" + (await coro(*args, **kwargs)) + "<after>"
+
+@decorator
+def coro_to_func(coro, *args, **kw):
+    return get_event_loop().run_until_complete(coro(*args, **kw))
+
+class CoroutineTestCase(unittest.TestCase):
+    def test_before_after(self):
+        @before_after
+        async def coro(x):
+           return x
+        self.assertTrue(inspect.iscoroutinefunction(coro))
+        out = get_event_loop().run_until_complete(coro('x'))
+        self.assertEqual(out, '<before>x<after>')
+
+    def test_coro_to_func(self):
+        @coro_to_func
+        async def coro(x):
+            return x
+        self.assertFalse(inspect.iscoroutinefunction(coro))
+        self.assertEqual(coro('x'), 'x')
+''')
+
+
 class DocumentationTestCase(unittest.TestCase):
     def test(self):
         err = doctest.testmod(doc)[0]
@@ -74,10 +108,40 @@ class ExtraTestCase(unittest.TestCase):
         @d1
         def f1(x, y, z):
             pass
-
         self.assertNotEqual(d1.__code__.co_filename, d2.__code__.co_filename)
         self.assertNotEqual(f1.__code__.co_filename, f2.__code__.co_filename)
-        self.assertNotEqual(f1_orig.__code__.co_filename, f1.__code__.co_filename)
+        self.assertNotEqual(f1_orig.__code__.co_filename,
+                            f1.__code__.co_filename)
+
+    def test_no_first_arg(self):
+        @decorator
+        def example(*args, **kw):
+            return args[0](*args[1:], **kw)
+
+        @example
+        def func(**kw):
+            return kw
+
+        # there is no confusion when passing args as a keyword argument
+        self.assertEqual(func(args='a'), {'args': 'a'})
+
+    def test_decorator_factory(self):
+        # similar to what IPython is doing in traitlets.config.application
+        @decorator
+        def catch_config_error(method, app, *args, **kwargs):
+            return method(app)
+        catch_config_error(lambda app: None)
+
+    def test_add1(self):
+        # similar to what IPython is doing in traitlets.config.application
+        @decorator
+        def add(func, const=1, *args, **kwargs):
+            return const + func(*args, **kwargs)
+
+        def f(x):
+            return x
+        self.assertEqual(add(2, f)(0), 2)
+
 
 # ################### test dispatch_on ############################# #
 # adapted from test_functools in Python 3.5
@@ -179,7 +243,6 @@ class TestSingleDispatch(unittest.TestCase):
         self.assertEqual(g(rnd), ("Number got rounded",))
 
     def test_register_abc(self):
-        c = collections
         d = {"a": "b"}
         l = [1, 2, 3]
         s = set([object(), None])
@@ -290,8 +353,6 @@ class TestSingleDispatch(unittest.TestCase):
         self.assertEqual(g(t), "tuple")
 
     def test_mro_conflicts(self):
-        c = collections
-
         @singledispatch
         def g(obj):
             return "base"
@@ -352,9 +413,9 @@ class TestSingleDispatch(unittest.TestCase):
         # MutableMapping's bases implicit as well from defaultdict's
         # perspective.
         with assertRaises(RuntimeError):
-            self.assertEqual(h(c.defaultdict(lambda: 0)), "sized")
+            self.assertEqual(h(defaultdict(lambda: 0)), "sized")
 
-        class R(c.defaultdict):
+        class R(defaultdict):
             pass
         c.MutableSequence.register(R)
 
@@ -397,5 +458,6 @@ class TestSingleDispatch(unittest.TestCase):
         with assertRaises(RuntimeError):
             h(u)
 
+
 if __name__ == '__main__':
     unittest.main()
